@@ -1,22 +1,17 @@
 """datajudge allows to assess  whether data from database complies with reference
 information."""
 
-
-import pkg_resources
-
 from .constraints.base import Constraint
-from .db_access import Condition
+from .db_access import Condition, DataSource
 from .requirements import BetweenRequirement, Requirement, WithinRequirement
 
 __all__ = [
     "BetweenRequirement",
     "Condition",
     "Constraint",
+    "DataSource",
     "Requirement",
     "WithinRequirement",
 ]
 
-try:
-    __version__ = pkg_resources.get_distribution(__name__).version
-except Exception:
-    __version__ = "1.5.0"
+__version__ = "1.11.0"
