@@ -1,18 +1,25 @@
+from __future__ import annotations
+
 import abc
+from collections.abc import Callable, Collection
 from dataclasses import dataclass, field
 from functools import lru_cache
-from typing import Any, Callable, List, Optional, Tuple, TypeVar
+from typing import Any, TypeVar
 
 import sqlalchemy as sa
 
 from ..db_access import DataReference
+from ..formatter import Formatter
+from ..utils import OutputProcessor, output_processor_limit
+
+DEFAULT_FORMATTER = Formatter()
 
 T = TypeVar("T")
-OptionalSelections = Optional[List[sa.sql.expression.Select]]
+OptionalSelections = list[sa.sql.expression.Select] | None
 ToleranceGetter = Callable[[sa.engine.Engine], float]
 
 
-def uncommon_substrings(string1: str, string2: str) -> Tuple[str, str]:
+def uncommon_substrings(string1: str, string2: str) -> tuple[str, str]:
     qualifiers1 = string1.split(".")
     qualifiers2 = string2.split(".")
     if qualifiers1[0] != qualifiers2[0]:
@@ -25,10 +32,30 @@ def uncommon_substrings(string1: str, string2: str) -> Tuple[str, str]:
 @dataclass(frozen=True)
 class TestResult:
     outcome: bool
-    failure_message: Optional[str] = field(default=None, repr=False)
-    constraint_description: Optional[str] = field(default=None, repr=False)
-    _factual_queries: Optional[str] = field(default=None, repr=False)
-    _target_queries: Optional[str] = field(default=None, repr=False)
+    _failure_message: str | None = field(default=None, repr=False)
+    _constraint_description: str | None = field(default=None, repr=False)
+    _factual_queries: str | None = field(default=None, repr=False)
+    _target_queries: str | None = field(default=None, repr=False)
+
+    def formatted_failure_message(self, formatter: Formatter) -> str | None:
+        return (
+            formatter.fmt_str(self._failure_message) if self._failure_message else None
+        )
+
+    def formatted_constraint_description(self, formatter: Formatter) -> str | None:
+        return (
+            formatter.fmt_str(self._constraint_description)
+            if self._constraint_description
+            else None
+        )
+
+    @property
+    def failure_message(self) -> str | None:
+        return self.formatted_failure_message(DEFAULT_FORMATTER)
+
+    @property
+    def constraint_description(self) -> str | None:
+        return self.formatted_constraint_description(DEFAULT_FORMATTER)
 
     @property
     def logging_message(self):
@@ -75,22 +102,35 @@ class Constraint(abc.ABC):
     """Express a DataReference constraint against either another DataReference or a reference value.
 
     Constraints against other DataReferences are typically referred to as 'between' constraints.
-    Please use the the `ref2` argument to instantiate such a constraint.
+    Please use the the ``ref2`` argument to instantiate such a constraint.
     Constraints against a fixed reference value are typically referred to as 'within' constraints.
-    Please use the `ref_value` argument to instantiate such a constraint.
+    Please use the ``ref_value`` argument to instantiate such a constraint.
 
     A constraint typically relies on the comparison of factual and target values. The former
     represent the key quantity of interest as seen in the database, the latter the key quantity of
     interest as expected a priori. Such a comparison is meant to be carried out in the `test`
     method.
 
-    In order to obtain such values, the `retrieve` method defines a mapping from DataReference,
-    be it the DataReference of primary interest, `ref`, or a baseline DataReference, `ref2`, to
-    value. If `ref_value` is already provided, usually no further mapping needs to be taken care of.
+    In order to obtain such values, the ``retrieve`` method defines a mapping from DataReference,
+    be it the DataReference of primary interest, ``ref``, or a baseline DataReference, ``ref2``, to
+    value. If ``ref_value`` is already provided, usually no further mapping needs to be taken care of.
+
+    By default, retrieved arguments are cached indefinitely ``@lru_cache(maxsize=None)``.
+    This can be controlled by setting the `cache_size` argument to a different value.
+    ``0`` disables caching.
     """
 
     def __init__(
-        self, ref: DataReference, *, ref2=None, ref_value: Any = None, name: str = None
+        self,
+        ref: DataReference,
+        *,
+        ref2: DataReference | None = None,
+        ref_value: Any = None,
+        name: str | None = None,
+        output_processors: OutputProcessor
+        | list[OutputProcessor]
+        | None = output_processor_limit,
+        cache_size=None,
     ):
         self._check_if_valid_between_or_within(ref2, ref_value)
         self.ref = ref
@@ -99,11 +139,29 @@ class Constraint(abc.ABC):
         self.name = name
         self.factual_selections: OptionalSelections = None
         self.target_selections: OptionalSelections = None
-        self.factual_queries: Optional[List[str]] = None
-        self.target_queries: Optional[List[str]] = None
+        self.factual_queries: list[str] | None = None
+        self.target_queries: list[str] | None = None
+
+        if (output_processors is not None) and (
+            not isinstance(output_processors, list)
+        ):
+            output_processors = [output_processors]
+        self.output_processors = output_processors
+
+        self.cache_size = cache_size
+        self._setup_caching()
+
+    def _setup_caching(self):
+        # this has an added benefit of allowing the class to be garbage collected
+        # according to https://rednafi.com/python/lru_cache_on_methods/
+        # and https://docs.astral.sh/ruff/rules/cached-instance-method/
+        self.get_factual_value = lru_cache(self.cache_size)(self.get_factual_value)  # type: ignore[method-assign]
+        self.get_target_value = lru_cache(self.cache_size)(self.get_target_value)  # type: ignore[method-assign]
 
     def _check_if_valid_between_or_within(
-        self, ref2: Optional[DataReference], ref_value: Optional[Any]
+        self,
+        ref2: DataReference | None,
+        ref_value: Any,
     ):
         """Check whether exactly one of ref2 and ref_value arguments have been used."""
         class_name = self.__class__.__name__
@@ -118,13 +176,13 @@ class Constraint(abc.ABC):
                 f"{class_name}. Use exactly either of them."
             )
 
-    @lru_cache(maxsize=None)
+    # @lru_cache(maxsize=None), see _setup_caching()
     def get_factual_value(self, engine: sa.engine.Engine) -> Any:
         factual_value, factual_selections = self.retrieve(engine, self.ref)
         self.factual_selections = factual_selections
         return factual_value
 
-    @lru_cache(maxsize=None)
+    # @lru_cache(maxsize=None), see _setup_caching()
     def get_target_value(self, engine: sa.engine.Engine) -> Any:
         if self.ref2 is None:
             return self.ref_value
@@ -149,7 +207,7 @@ class Constraint(abc.ABC):
 
     @property
     def target_prefix(self) -> str:
-        return f"{self.ref2.get_string()}'s " if (self.ref2 is not None) else ""
+        return f"{self.ref2}'s " if (self.ref2 is not None) else ""
 
     @property
     def condition_string(self) -> str:
@@ -175,13 +233,11 @@ class Constraint(abc.ABC):
 
     def retrieve(
         self, engine: sa.engine.Engine, ref: DataReference
-    ) -> Tuple[Any, OptionalSelections]:
+    ) -> tuple[Any, OptionalSelections]:
         """Retrieve the value of interest for a DataReference from database."""
         pass
 
-    def compare(
-        self, value_factual: Any, value_target: Any
-    ) -> Tuple[bool, Optional[str]]:
+    def compare(self, value_factual: Any, value_target: Any) -> tuple[bool, str | None]:
         pass
 
     def test(self, engine: sa.engine.Engine) -> TestResult:
@@ -218,6 +274,12 @@ class Constraint(abc.ABC):
             target_queries,
         )
 
+    def apply_output_formatting(self, values: Collection) -> Collection:
+        if self.output_processors is not None:
+            for output_processor in self.output_processors:
+                values, _ = output_processor(values)
+        return values
+
 
 def format_sample(sample, ref: DataReference) -> str:
     """Build a string from a database row indicating its column values."""
