@@ -1,6 +1,7 @@
+from __future__ import annotations
+
 import abc
-from functools import lru_cache
-from typing import List, Optional, Tuple
+from functools import cache
 
 import sqlalchemy as sa
 
@@ -15,12 +16,15 @@ class Row(Constraint, abc.ABC):
         ref: DataReference,
         ref2: DataReference,
         max_missing_fraction_getter: ToleranceGetter,
-        name: str = None,
+        name: str | None = None,
+        cache_size=None,
     ):
-        super().__init__(ref, ref2=ref2, name=name)
+        super().__init__(ref, ref2=ref2, name=name, cache_size=cache_size)
         self.max_missing_fraction_getter = max_missing_fraction_getter
 
     def test(self, engine: sa.engine.Engine) -> TestResult:
+        if self.ref is None or self.ref2 is None:
+            raise ValueError()
         if db_access.is_impala(engine):
             raise NotImplementedError("Currently not implemented for impala.")
         self.max_missing_fraction = self.max_missing_fraction_getter(engine)
@@ -34,8 +38,9 @@ class Row(Constraint, abc.ABC):
 
 
 class RowEquality(Row):
-    @lru_cache(maxsize=None)
-    def get_factual_value(self, engine: sa.engine.Engine) -> Tuple[int, int]:
+    def get_factual_value(self, engine: sa.engine.Engine) -> tuple[int, int]:
+        if self.ref is None or self.ref2 is None:
+            raise ValueError()
         n_rows_missing_left, selections_left = db_access.get_row_difference_count(
             engine, self.ref, self.ref2
         )
@@ -45,8 +50,9 @@ class RowEquality(Row):
         self.factual_selections = [*selections_left, *selections_right]
         return n_rows_missing_left, n_rows_missing_right
 
-    @lru_cache(maxsize=None)
     def get_target_value(self, engine: sa.engine.Engine) -> int:
+        if self.ref is None or self.ref2 is None:
+            raise ValueError()
         n_rows_total, selections = db_access.get_unique_count_union(
             engine, self.ref, self.ref2
         )
@@ -55,13 +61,15 @@ class RowEquality(Row):
 
     # fraction: (|T1 - T2| + |T2 - T1|) / |T1 U T2|
     def compare(
-        self, n_rows_missing_tuple: Tuple[int, int], n_rows_total: int
-    ) -> Tuple[bool, Optional[str]]:
+        self, n_rows_missing_tuple: tuple[int, int], n_rows_total: int
+    ) -> tuple[bool, str | None]:
         n_rows_missing_left, n_rows_missing_right = n_rows_missing_tuple
         missing_fraction = (n_rows_missing_left + n_rows_missing_right) / n_rows_total
         result = missing_fraction <= self.max_missing_fraction
         if result:
             return result, None
+        if self.ref2 is None:
+            raise ValueError("RowEquality constraint requires ref2.")
         if n_rows_missing_left > 0:
             sample_string = format_sample(self.ref1_minus_ref2_sample, self.ref2)
         else:
@@ -69,16 +77,18 @@ class RowEquality(Row):
         assertion_message = (
             f"{missing_fraction} > "
             f"{self.max_missing_fraction} of rows differ "
-            f"between {self.ref.get_string()} and "
-            f"{self.ref2.get_string()}. E.g. for "
+            f"between {self.ref} and "
+            f"{self.ref2}. E.g. for "
             f"{sample_string}."
         )
         return result, assertion_message
 
 
 class RowSubset(Row):
-    @lru_cache(maxsize=None)
+    @cache
     def get_factual_value(self, engine: sa.engine.Engine) -> int:
+        if self.ref is None or self.ref2 is None:
+            raise ValueError()
         n_rows_missing, selections = db_access.get_row_difference_count(
             engine,
             self.ref,
@@ -87,16 +97,16 @@ class RowSubset(Row):
         self.factual_selections = selections
         return n_rows_missing
 
-    @lru_cache(maxsize=None)
+    @cache
     def get_target_value(self, engine: sa.engine.Engine) -> int:
         n_rows_total, selections = db_access.get_unique_count(engine, self.ref)
         self.target_selections = selections
         return n_rows_total
 
-    @lru_cache(maxsize=None)
+    @cache
     def compare(
         self, n_rows_missing: int, n_rows_total: int
-    ) -> Tuple[bool, Optional[str]]:
+    ) -> tuple[bool, str | None]:
         if n_rows_total == 0:
             return True, None
         missing_fraction = n_rows_missing / n_rows_total
@@ -107,8 +117,8 @@ class RowSubset(Row):
         assertion_message = (
             f"{missing_fraction} > "
             f"{self.max_missing_fraction} of rows of "
-            f"{self.ref.get_string()} are "
-            f"not in {self.ref2.get_string()}. E.g. for "
+            f"{self.ref} are "
+            f"not in {self.ref2}. E.g. for "
             f"{sample_string}. "
             f"{self.condition_string} "
         )
@@ -116,35 +126,39 @@ class RowSubset(Row):
 
 
 class RowSuperset(Row):
-    @lru_cache(maxsize=None)
     def get_factual_value(self, engine: sa.engine.Engine) -> int:
+        if self.ref is None or self.ref2 is None:
+            raise ValueError()
         n_rows_missing, selections = db_access.get_row_difference_count(
             engine, self.ref2, self.ref
         )
         self.factual_selections = selections
         return n_rows_missing
 
-    @lru_cache(maxsize=None)
     def get_target_value(self, engine: sa.engine.Engine) -> int:
+        if self.ref is None or self.ref2 is None:
+            raise ValueError()
         n_rows_total, selections = db_access.get_unique_count(engine, self.ref2)
         self.target_selections = selections
         return n_rows_total
 
     def compare(
         self, n_rows_missing: int, n_rows_total: int
-    ) -> Tuple[bool, Optional[str]]:
+    ) -> tuple[bool, str | None]:
         if n_rows_total == 0:
             return True, None
         missing_fraction = n_rows_missing / n_rows_total
         result = missing_fraction <= self.max_missing_fraction
         if result:
             return result, None
+        if self.ref2 is None:
+            raise ValueError("RowSuperset constraint requires ref2.")
         sample_string = format_sample(self.ref2_minus_ref1_sample, self.ref2)
         assertion_message = (
             f"{missing_fraction} > "
             f"{self.max_missing_fraction} of rows of "
-            f"{self.ref2.get_string()} are "
-            f"not in {self.ref.get_string()}. E.g. for "
+            f"{self.ref2} are "
+            f"not in {self.ref}. E.g. for "
             f"{sample_string}. "
             f"{self.condition_string} "
         )
@@ -156,18 +170,20 @@ class RowMatchingEquality(Row):
         self,
         ref: DataReference,
         ref2: DataReference,
-        matching_columns1: List[str],
-        matching_columns2: List[str],
-        comparison_columns1: List[str],
-        comparison_columns2: List[str],
+        matching_columns1: list[str],
+        matching_columns2: list[str],
+        comparison_columns1: list[str],
+        comparison_columns2: list[str],
         max_missing_fraction_getter: ToleranceGetter,
-        name: str = None,
+        name: str | None = None,
+        cache_size=None,
     ):
         super().__init__(
             ref,
             ref2=ref2,
             max_missing_fraction_getter=max_missing_fraction_getter,
             name=name,
+            cache_size=cache_size,
         )
         self.match_and_compare = MatchAndCompare(
             matching_columns1,
@@ -177,6 +193,8 @@ class RowMatchingEquality(Row):
         )
 
     def test(self, engine: sa.engine.Engine) -> TestResult:
+        if self.ref is None or self.ref2 is None:
+            raise ValueError()
         missing_fraction, n_rows_match, selections = db_access.get_row_mismatch(
             engine, self.ref, self.ref2, self.match_and_compare
         )
@@ -188,8 +206,8 @@ class RowMatchingEquality(Row):
         assertion_message = (
             f"{missing_fraction} > "
             f"{max_missing_fraction} of the rows differ "
-            f"on a match of {n_rows_match} rows between {self.ref.get_string()} and "
-            f"{self.ref2.get_string()}. "
+            f"on a match of {n_rows_match} rows between {self.ref} and "
+            f"{self.ref2}. "
             f"{self.condition_string}"
             f"{self.match_and_compare} "
         )
