@@ -1,6 +1,7 @@
+from __future__ import annotations
+
 import math
 import warnings
-from typing import List, Optional, Tuple
 
 import sqlalchemy as sa
 
@@ -15,15 +16,14 @@ class KolmogorovSmirnov2Sample(Constraint):
         ref: DataReference,
         ref2: DataReference,
         significance_level: float = 0.05,
-        name: str = None,
+        name: str | None = None,
+        cache_size=None,
     ):
         self.significance_level = significance_level
-        super().__init__(ref, ref2=ref2, name=name)
+        super().__init__(ref, ref2=ref2, name=name, cache_size=cache_size)
 
     @staticmethod
-    def approximate_p_value(
-        d: float, n_samples: int, m_samples: int
-    ) -> Optional[float]:
+    def approximate_p_value(d: float, n_samples: int, m_samples: int) -> float | None:
         """
         Calculates the approximate p-value according to
         'A procedure to find exact critical values of Kolmogorov-Smirnov Test', Silvia Fachinetti, 2009
@@ -74,10 +74,10 @@ class KolmogorovSmirnov2Sample(Constraint):
 
     @staticmethod
     def calculate_statistic(
-        engine,
+        engine: sa.engine.Engine,
         ref1: DataReference,
         ref2: DataReference,
-    ) -> Tuple[float, Optional[float], int, int, List]:
+    ) -> tuple[float, float | None, int, int, list[sa.Select]]:
         # retrieve test statistic d, as well as sample sizes m and n
         d_statistic, ks_selections = db_access.get_ks_2sample(
             engine,
@@ -97,6 +97,8 @@ class KolmogorovSmirnov2Sample(Constraint):
         return d_statistic, p_value, n_samples, m_samples, selections
 
     def test(self, engine: sa.engine.Engine) -> TestResult:
+        if self.ref2 is None:
+            raise ValueError("KolmogorovSmirnov2Sample requires ref2.")
         (
             d_statistic,
             p_value,
@@ -114,7 +116,7 @@ class KolmogorovSmirnov2Sample(Constraint):
 
         assertion_text = (
             f"Null hypothesis (H0) for the 2-sample Kolmogorov-Smirnov test was rejected, i.e., "
-            f"the two samples ({self.ref.get_string()} and {self.target_prefix}) "
+            f"the two samples ({self.ref} and {self.target_prefix}) "
             f"do not originate from the same distribution. "
             f"The test results are d={d_statistic}"
         )
