@@ -6,7 +6,6 @@ import urllib.parse
 
 import pytest
 import sqlalchemy as sa
-from impala.dbapi import connect
 
 from datajudge.db_access import apply_patches, is_bigquery, is_db2, is_impala, is_mssql
 
@@ -18,6 +17,7 @@ def get_engine(backend) -> sa.engine.Engine:
     address = os.environ.get("DB_ADDR", "localhost")
 
     if backend == "impala":
+        from impala.dbapi import connect
 
         def conn_creator():
             return connect(
@@ -52,34 +52,36 @@ def get_engine(backend) -> sa.engine.Engine:
         # gcp_project = os.environ.get("GOOGLE_CLOUD_PROJECT", "scratch-361908")
         connection_string = "bigquery://"
 
-    engine = sa.create_engine(connection_string, echo=True)
+    engine = sa.create_engine(
+        connection_string, echo=True, pool_size=10, max_overflow=20
+    )
     apply_patches(engine)
 
     return engine
 
 
-def _string_column(engine):
+def _string_column(engine, minlength_db2=40):
     if is_db2(engine):
-        return sa.String(40)
+        return sa.String(minlength_db2)
     return sa.String()
 
 
 @pytest.fixture(scope="module")
 def engine(backend):
     engine = get_engine(backend)
-    with engine.connect() as conn:
+    with engine.begin() as conn:
         if engine.name in ("postgresql", "bigquery", "impala"):
-            conn.execute(f"CREATE SCHEMA IF NOT EXISTS {SCHEMA}")
+            conn.execute(sa.text(f"CREATE SCHEMA IF NOT EXISTS {SCHEMA}"))
     return engine
 
 
-@pytest.fixture(scope="module")
+@pytest.fixture(scope="session")
 def metadata():
     return sa.MetaData()
 
 
 def _handle_table(engine, metadata, table_name, columns, data):
-    with engine.connect() as conn:
+    with engine.begin() as conn:
         if sa.inspect(conn).has_table(table_name, schema=SCHEMA):
             return
         table = sa.Table(table_name, metadata, *columns, schema=SCHEMA)
@@ -262,6 +264,19 @@ def date_table_overlap(engine, metadata):
             "date_end": datetime.datetime(2016, 1, 8),
         },
     ]
+    # Duplicate entries.
+    data += [
+        {
+            "id1": 6,
+            "date_start": datetime.datetime(2016, 1, 1),
+            "date_end": datetime.datetime(2016, 1, 10),
+        },
+        {
+            "id1": 6,
+            "date_start": datetime.datetime(2016, 1, 1),
+            "date_end": datetime.datetime(2016, 1, 10),
+        },
+    ]
     _handle_table(engine, metadata, table_name, columns, data)
     return TEST_DB_NAME, SCHEMA, table_name
 
@@ -379,6 +394,105 @@ def date_table_overlap_2d(engine, metadata):
             "date_end2": datetime.datetime(2017, 1, 7),
         },
     ]
+    # Duplication overlap.
+    data += [
+        {
+            "id1": 8,
+            "date_start1": datetime.datetime(2016, 1, 1),
+            "date_end1": datetime.datetime(2016, 1, 10),
+            "date_start2": datetime.datetime(2017, 1, 1),
+            "date_end2": datetime.datetime(2017, 1, 10),
+        },
+        {
+            "id1": 8,
+            "date_start1": datetime.datetime(2016, 1, 1),
+            "date_end1": datetime.datetime(2016, 1, 10),
+            "date_start2": datetime.datetime(2017, 1, 1),
+            "date_end2": datetime.datetime(2017, 1, 10),
+        },
+    ]
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
+@pytest.fixture(scope="module")
+def integer_table_overlap(engine, metadata):
+    table_name = "integer_table_overlap"
+    columns = [
+        sa.Column("id1", sa.Integer()),
+        sa.Column("range_start", sa.Integer()),
+        sa.Column("range_end", sa.Integer()),
+    ]
+    data = []
+    # Trivial case: single entry.
+    data += [
+        {
+            "id1": 1,
+            "range_start": 1,
+            "range_end": 10,
+        }
+    ]
+    # 'Normal case': Multiple entries without overlap.
+    data += [
+        {
+            "id1": 2,
+            "range_start": i * 2,
+            "range_end": i * 2 + 1,
+        }
+        for i in range(1, 5)
+    ]
+    # Multiple entries with non-singleton overlap.
+    data += [
+        {
+            "id1": 3,
+            "range_start": 1,
+            "range_end": 10,
+        },
+        {
+            "id1": 3,
+            "range_start": 7,
+            "range_end": 15,
+        },
+    ]
+    # Multiple entries with singleton overlap.
+    data += [
+        {
+            "id1": 4,
+            "range_start": 1,
+            "range_end": 10,
+        },
+        {
+            "id1": 4,
+            "range_start": 10,
+            "range_end": 15,
+        },
+    ]
+    # Multiple entries with subset relation.
+    data += [
+        {
+            "id1": 5,
+            "range_start": 1,
+            "range_end": 10,
+        },
+        {
+            "id1": 5,
+            "range_start": 4,
+            "range_end": 8,
+        },
+    ]
+    # Duplicate entries.
+    data += [
+        {
+            "id1": 6,
+            "range_start": 1,
+            "range_end": 10,
+        },
+        {
+            "id1": 6,
+            "range_start": 1,
+            "range_end": 10,
+        },
+    ]
     _handle_table(engine, metadata, table_name, columns, data)
     return TEST_DB_NAME, SCHEMA, table_name
 
@@ -452,6 +566,131 @@ def date_table_gap(engine, metadata):
     return TEST_DB_NAME, SCHEMA, table_name
 
 
+@pytest.fixture(scope="module")
+def integer_table_gap(engine, metadata):
+    table_name = "integer_table_gap"
+    columns = [
+        sa.Column("id1", sa.Integer()),
+        sa.Column("range_start", sa.Integer()),
+        sa.Column("range_end", sa.Integer()),
+    ]
+    data = []
+    # Single entry should not be considered a gap.
+    data += [
+        {
+            "id1": 1,
+            "range_start": 1,
+            "range_end": 3,
+        }
+    ]
+    # Multiple entries without gap.
+    data += [
+        {
+            "id1": 2,
+            "range_start": 3 + i * 2,
+            "range_end": 5 + i * 2,
+        }
+        for i in range(1, 5)
+    ]
+    # Multiple entries with overlap.
+    data += [
+        {
+            "id1": 3,
+            "range_start": 1,
+            "range_end": 10,
+        },
+        {
+            "id1": 3,
+            "range_start": 3,
+            "range_end": 7,
+        },
+    ]
+    # Multiple entries with gap.
+    data += [
+        {
+            "id1": 4,
+            "range_start": 1,
+            "range_end": 5,
+        },
+        {
+            "id1": 4,
+            "range_start": 7,
+            "range_end": 10,
+        },
+    ]
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
+@pytest.fixture(scope="module")
+def float_table_gap(engine, metadata):
+    table_name = "float_table_gap"
+    columns = [
+        sa.Column("id1", sa.Integer()),
+        sa.Column("range_start", sa.Float()),
+        sa.Column("range_end", sa.Float()),
+    ]
+    data: list[dict[str, int | float]] = []
+    # Single entry should not be considered a gap.
+    data += [
+        {
+            "id1": 1,
+            "range_start": 1,
+            "range_end": 3,
+        }
+    ]
+    # Multiple entries without gap.
+    data += [
+        {
+            "id1": 2,
+            "range_start": 3 + i * 2,
+            "range_end": 5 + i * 2,
+        }
+        for i in range(1, 5)
+    ]
+    # Multiple entries with overlap.
+    data += [
+        {
+            "id1": 3,
+            "range_start": 1,
+            "range_end": 10,
+        },
+        {
+            "id1": 3,
+            "range_start": 3,
+            "range_end": 7,
+        },
+    ]
+    # Multiple entries with gap.
+    data += [
+        {
+            "id1": 4,
+            "range_start": 1,
+            "range_end": 5,
+        },
+        {
+            "id1": 4,
+            "range_start": 8,
+            "range_end": 10,
+        },
+    ]
+    # Multiple entries with tolerated gap.
+    data += [
+        {
+            "id1": 5,
+            "range_start": 1,
+            "range_end": 5,
+        },
+        {
+            "id1": 5,
+            "range_start": 5.5,
+            "range_end": 10,
+        },
+    ]
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
 @pytest.fixture(scope="module")
 def date_table_keys(engine, metadata):
     table_name = "date_table_keys"
@@ -480,6 +719,18 @@ def date_table_keys(engine, metadata):
     return TEST_DB_NAME, SCHEMA, table_name
 
 
+@pytest.fixture(scope="module")
+def distribution_table(engine, metadata):
+    table_name = "distribution_table"
+    columns = [
+        sa.Column("col_int", sa.Integer()),
+        sa.Column("col_varchar", _string_column(engine)),
+    ]
+    data = [{"col_int": i // 10, "col_varchar": f"hi{i // 15}"} for i in range(20)]
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
 @pytest.fixture(scope="module")
 def unique_table1(engine, metadata):
     table_name = "unique_table1"
@@ -508,10 +759,54 @@ def unique_table2(engine, metadata):
     return TEST_DB_NAME, SCHEMA, table_name
 
 
+@pytest.fixture(scope="module")
+def unique_table_extralong(engine, metadata):
+    if is_impala(engine):
+        pytest.skip(
+            "Skipping this larger output check for impala due to it being quite brittle"
+        )
+    if is_bigquery(engine):
+        pytest.skip(
+            "Skipping this larger output check for bigquery since creating the table is very slow"
+        )
+    table_name = "unique_table_extralong"
+    columns = [
+        sa.Column("col_int", sa.Integer()),
+        sa.Column("col_varchar", _string_column(engine)),
+    ]
+    data = [{"col_int": i // 2, "col_varchar": f"hi{i // 3}"} for i in range(12345)]
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
+@pytest.fixture(scope="module")
+def unique_table_largesize(engine, metadata):
+    if is_impala(engine):
+        pytest.skip(
+            "Skipping this larger output check for impala due to it being quite brittle"
+        )
+    if is_bigquery(engine):
+        pytest.skip(
+            "Skipping this larger output check for bigquery since creating the table is very slow"
+        )
+    table_name = "unique_table_largesize"
+    columns = [
+        sa.Column("col_int", sa.Integer()),
+        sa.Column("col_varchar", _string_column(engine, minlength_db2=2000)),
+    ]
+    data = [
+        {"col_int": i // 2, "col_varchar": "hi" * min(i, 900)} for i in range(12345)
+    ]
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
 @pytest.fixture(scope="module")
 def nested_table(engine, metadata):
     table_name = "nested_table"
-    columns = [sa.Column("nested_varchar", _string_column(engine))]
+    columns: list[sa.Column | str] = [
+        sa.Column("nested_varchar", _string_column(engine))
+    ]
     data = [
         {"nested_varchar": "ABC#1,"},
         {"nested_varchar": "ABC#1,DEF#2,"},
@@ -521,13 +816,85 @@ def nested_table(engine, metadata):
     return TEST_DB_NAME, SCHEMA, table_name
 
 
+@pytest.fixture(scope="module")
+def functional_dependency_table(engine, metadata):
+    table_name = "functional_dependency_table"
+
+    # a -> b
+    # c -> b
+    # a -/-> c
+    columns = [
+        sa.Column("a", sa.Integer()),
+        sa.Column("b", sa.Integer()),
+        sa.Column("c", sa.Integer()),
+    ]
+    data = [
+        {"a": 1, "b": 2, "c": 3},
+        {"a": 1, "b": 2, "c": 4},
+        {"a": 2, "b": 3, "c": 5},
+        {"a": 2, "b": 3, "c": 6},
+        {"a": 3, "b": 4, "c": 1},
+        {"a": 3, "b": 4, "c": 2},
+    ]
+
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
+@pytest.fixture(scope="module")
+def functional_dependency_table_multi_key(engine, metadata):
+    table_name = "functional_dependency_table_multi_key"
+
+    # ab -> c
+    # ab -/-> d
+    # ab -/-> ce
+    columns = [
+        sa.Column("a", sa.Integer()),
+        sa.Column("b", sa.Integer()),
+        sa.Column("c", sa.Integer()),
+        sa.Column("d", sa.Integer()),
+        sa.Column("e", sa.Integer()),
+    ]
+
+    # fmt: off
+    data = [
+        {"a": 1, "b": 1, "c": 2, "d": 3, "e": 2, },
+        {"a": 1, "b": 1, "c": 2, "d": 4, "e": 2, },
+        {"a": 1, "b": 2, "c": 3, "d": 5, "e": 3, },
+        {"a": 1, "b": 2, "c": 3, "d": 6, "e": 3, },
+        {"a": 2, "b": 1, "c": 4, "d": 7, "e": 4, },
+        {"a": 2, "b": 1, "c": 4, "d": 8, "e": 4, },
+        {"a": 2, "b": 2, "c": 5, "d": 9, "e": 5, },
+        {"a": 2, "b": 2, "c": 5, "d": 10, "e": 5, },
+
+        # if NULL is on the LHS, this is not considered a functional dependency violation
+        {"a": None, "b": None, "c": 6, "d": 10, "e": 6, },
+        {"a": None, "b": None, "c": 6, "d": 11, "e": 8, },
+
+        {"a": None, "b": 99, "c": 6, "d": 10, "e": 6, },
+        {"a": None, "b": 99, "c": 6, "d": 11, "e": 8, },
+        {"a": 42, "b": None, "c": 6, "d": 11, "e": 6, },
+        {"a": None, "b": 42, "c": 6, "d": 11, "e": 6, },
+        {"a": 43, "b": 43, "c": 6, "d": 12, "e": 6, },
+        {"a": 43, "b": 43, "c": 6, "d": 12, "e": 7, },
+        {"a": 44, "b": 44, "c": None, "d": 12, "e": None, },
+        {"a": 44, "b": 44, "c": None, "d": 13, "e": 99, },
+    ]
+    # fmt: on
+
+    _handle_table(engine, metadata, table_name, columns, data)
+    return TEST_DB_NAME, SCHEMA, table_name
+
+
 @pytest.fixture(scope="module")
 def varchar_table1(engine, metadata):
     table_name = "varchar_table1"
-    columns = [
+    columns: list[sa.Column | str] = [
         sa.Column("col_varchar", _string_column(engine)),
     ]
-    data = [{"col_varchar": "qq" * i} for i in range(1, 10)]
+    data: list[dict[str, str | None]] = [
+        {"col_varchar": "qq" * i} for i in range(1, 10)
+    ]
     data.append({"col_varchar": None})
     _handle_table(engine, metadata, table_name, columns, data)
     return TEST_DB_NAME, SCHEMA, table_name
@@ -536,7 +903,7 @@ def varchar_table1(engine, metadata):
 @pytest.fixture(scope="module")
 def varchar_table2(engine, metadata):
     table_name = "varchar_table2"
-    columns = [
+    columns: list[sa.Column | str] = [
         sa.Column("col_varchar", _string_column(engine)),
     ]
     data = [{"col_varchar": "qq" * i} for i in range(2, 11)]
@@ -547,7 +914,7 @@ def varchar_table2(engine, metadata):
 @pytest.fixture(scope="module")
 def varchar_table_real(engine, metadata):
     table_name = "varchar_table_real"
-    columns = [
+    columns: list[sa.Column | str] = [
         sa.Column("col_varchar", _string_column(engine)),
     ]
     data = [
@@ -671,10 +1038,10 @@ def groupby_aggregation_table_correct(engine, metadata):
         sa.Column("value", sa.Integer()),
     ]
     data_source = [
-        [34807101, 8, {1}],
-        [42760071, 3, {7, 3, 4, 5, 6, 1, 2}],
-        [42760071, 7, {1}],
-        [44093821, 10, {2, 8, 3, 6, 7, 4, 1, 5}],
+        (34807101, 8, {1}),
+        (42760071, 3, {7, 3, 4, 5, 6, 1, 2}),
+        (42760071, 7, {1}),
+        (44093821, 10, {2, 8, 3, 6, 7, 4, 1, 5}),
     ]
     data = [
         {"some_id": id, "extra_id": e, "value": v}
@@ -694,10 +1061,10 @@ def groupby_aggregation_table_incorrect(engine, metadata):
         sa.Column("value", sa.Integer()),
     ]
     data_source = [
-        [34807101, 8, {1}],
-        [42760071, 3, {7, 3, 4, 5, 6, 1, 2}],
-        [42760071, 7, {1}],
-        [44093821, 7, {22, 19, 16, 24, 23, 21, 20, 18, 17}],
+        (34807101, 8, {1}),
+        (42760071, 3, {7, 3, 4, 5, 6, 1, 2}),
+        (42760071, 7, {1}),
+        (44093821, 7, {22, 19, 16, 24, 23, 21, 20, 18, 17}),
     ]
     data = [
         {"some_id": id, "extra_id": e, "value": v}
@@ -768,16 +1135,23 @@ def capitalization_table(engine, metadata):
         primary_key = ""
     else:
         str_datatype = "TEXT"
-    with engine.connect() as connection:
-        connection.execute(f"DROP TABLE IF EXISTS {SCHEMA}.{table_name}")
+
+    with engine.begin() as connection:
+        if sa.inspect(connection).has_table(table_name, schema=SCHEMA):
+            return TEST_DB_NAME, SCHEMA, table_name, uppercase_column, lowercase_column
+        connection.execute(sa.text(f"DROP TABLE IF EXISTS {SCHEMA}.{table_name}"))
         connection.execute(
-            f"CREATE TABLE {SCHEMA}.{table_name} "
-            f"(id INTEGER {primary_key}, "
-            f"{uppercase_column} {str_datatype}, {lowercase_column} INTEGER)"
+            sa.text(
+                f"CREATE TABLE {SCHEMA}.{table_name} "
+                f"(id INTEGER {primary_key}, "
+                f"{uppercase_column} {str_datatype}, {lowercase_column} INTEGER)"
+            )
         )
         connection.execute(
-            f"INSERT INTO {SCHEMA}.{table_name} "
-            f"(id, {uppercase_column}, {lowercase_column}) VALUES (1, 'QuantCo', 100)"
+            sa.text(
+                f"INSERT INTO {SCHEMA}.{table_name} "
+                f"(id, {uppercase_column}, {lowercase_column}) VALUES (1, 'QuantCo', 100)"
+            )
         )
     return TEST_DB_NAME, SCHEMA, table_name, uppercase_column, lowercase_column
 
